/*	$Id$	*/

/*
 * Copyright 2018 Maverick Embedded Technology Ltd.
 * All Rights Reserved.
 *
 * Written by Steve Woodford.
 *
 * This software may only be used with devices supplied by Maverick
 * Embedded Technology Ltd. Use for any purpose other than downloading
 * firmware images supplied by Maverick Embedded Technology Ltd is
 * not permitted.
 */

/*
 * Firmware downloader for network-enabled embedded devices from
 * Maverick Embedded Technology Ltd.
 *
 * This source code should build on most modern operating systems
 * include MacOS, Linux, *BSD, and Windows.
 *
 * For non-Windows hosts:
 *
 *	$ cc -O2 maverick-dl.c -o maverick-dl
 *
 * For Windows, using Visual Studio 2017 (from the Developer Command Prompt):
 *
 *	C:\ cl /O2 maverick-dl.c /link ws2_32.lib
 *
 * Usage:
 *	maverick-dl <image-filename> <target-ip-address>
 *
 * Pre-compiled binaries are supplied for MacOS and Windows. The source is
 * provided both for reference (you can validate it to ensure it won't do
 * anything nefarious!) and for building on other operating systems.
 */

#ifndef _WIN32
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <netdb.h>
#include <err.h>
#include <unistd.h>
#else
#undef UNICODE
#include <ws2tcpip.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * The bootloader listens for network connections on this port.
 */
#define	BOOTLOADER_PORT	"8007"

/*
 * Maximum number of times we will re-send the same block to the
 * target.
 */
#define	MAX_RETRIES	5

/*
 * Blocks, as transmitted to the target.
 * A 32-bit little-endian sequence/block number is prepended to
 * each 512 bytes of data.
 */
#define	WAVR_BLOCK_SIZE	512
static uint8_t out_buff[sizeof(uint32_t) + WAVR_BLOCK_SIZE];

#ifndef _WIN32
#define	socket_fd_t	int
#define	closesocket	close
#define	local_err	err
#define local_errx	errx
#else
static const char *pname;
static const char *errstr;
#define	socket_fd_t	SOCKET

static void
local_err_common(const char *fmt, va_list args, int nl)
{

	fprintf(stderr, "%s: ", pname);
	vfprintf(stderr, fmt, args);

	if (nl)
		fputc('\n', stderr);
	else {
		fputc(':', stderr);
		fputc(' ', stderr);
	}

#ifdef _WIN32
	WSACleanup();
#endif
}

static void
local_errx(int exit_code, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	local_err_common(fmt, args, 1);
	va_end(args);
	exit(exit_code);
}

static void
local_err(int exit_code, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	local_err_common(fmt, args, 0);
	va_end(args);
	fprintf(stderr, "%s\n", errstr);
	exit(exit_code);
}
#endif /* _WIN32 */

static const char *
get_response(socket_fd_t fd)
{
	static char reply[WAVR_BLOCK_SIZE];
	unsigned int len;

	len = 0;

	/*
	 * Responses are always terminated by a nul character, so just
	 * keep reading until we see one or until the buffer overflows.
	 * The latter case indicates an error.
	 */
	while (len < sizeof(reply)) {
		if (recv(fd, &reply[len], 1, 0) != 1) {
#ifdef _WIN32
			errstr = strerror(WSAGetLastError());
#endif
			local_err(1, "read reply");
		}

		/*
		 * If, for whatever reason, we receive a whole slew of
		 * nul characters before anything useful, just skip.
		 */
		if (reply[len] == '\0') {
			if (len)
				return (reply);
		} else
			len += 1;
	}

	/* Overflow */
	return (NULL);
}

int
main(int argc, char **argv)
{
	struct addrinfo hints, *ai;
	struct stat st;
	uint32_t seq, expected_seq, max_seq;
	size_t img_size;
	uint8_t *img;
	int rv, retries;
	socket_fd_t fd;
	const char *resp;
	FILE *ifp;

	if (argc != 3) {
		fprintf(stderr, "usage: %s <image-file> <target-ip-or-name>\n",
		    argv[0]);
		exit(1);
	}

#ifdef _WIN32
	WSADATA wsaData;
	int i;

	i = WSAStartup(MAKEWORD(2,2), &wsaData);
	if (i != NO_ERROR) {
		fprintf(stderr, "WSAStartup failed with error: %d\n", i);
		exit(1);
	}
	pname = argv[0];
#endif

	/*
	 * Verify the firware image exists and is a real file
	 */
	if (stat(argv[1], &st) < 0) {
#ifdef _WIN32
		errstr = strerror(errno);
#endif
		local_err(1, "%s", argv[1]);
	}

#ifndef _WIN32
	if (!S_ISREG(st.st_mode))
		local_errx(1, "%s is not a plain file", argv[1]);
#endif

	/*
	 * Make sure the size is reasonable
	 */
	if (st.st_size < 0x10000)
		local_errx(1, "Image file must be at least 64 KB in size");

	if (st.st_size > 0x38000)
		local_errx(1, "Image file is greater than 192 KB in size");

	/*
	 * Round up the file size to a multiple of WAVR_BLOCK_SIZE
	 */
	img_size = st.st_size;
	if (img_size % WAVR_BLOCK_SIZE)
		img_size += WAVR_BLOCK_SIZE - (img_size % WAVR_BLOCK_SIZE);

	/*
	 * Add one extra block; the target may request an additional
	 * padding block.
	 */
	img_size += WAVR_BLOCK_SIZE;

	/*
	 * Allocate the staging buffer
	 */
	if ((img = calloc(1, img_size)) == NULL) {
#ifdef _WIN32
		errstr = strerror(errno);
#endif
		local_err(1, "malloc(%" PRIu64 ")", (uint64_t)img_size);
	}

	/*
	 * Open the firmware image and read it into the buffer
	 */
	if ((ifp = fopen(argv[1], "rb")) == NULL) {
#ifdef _WIN32
		errstr = strerror(errno);
#endif
		local_err(1, "fopen(%s)", argv[1]);
	}
	if (fread(img, 1, st.st_size, ifp) != (size_t)st.st_size) {
#ifdef _WIN32
		errstr = strerror(errno);
#endif
		local_err(1, "fread(%s, %" PRIu64 ")", argv[1],
		    (uint64_t)st.st_size);
	}
	fclose(ifp);

	/*
	 * Convert the supplied target IP/address into something
	 * we can pass to socket(2) and connect(2);
	 */
	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = 0;
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	rv = getaddrinfo(argv[2], BOOTLOADER_PORT, &hints, &ai);
	if (rv != 0)
		local_errx(1, "getaddrinfo(%s): %s", argv[2], gai_strerror(rv));

	fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
#ifndef _WIN32
	if (fd < 0)
		local_err(1, "socket");
#else
	if (fd == INVALID_SOCKET) {
		errstr = strerror(WSAGetLastError());
		local_err(1, "socket");
	}
#endif

	rv = connect(fd, ai->ai_addr, ai->ai_addrlen);
	if (rv < 0) {
#ifdef _WIN32
		errstr = strerror(WSAGetLastError());
#endif
		local_err(1, "connect");
	}

	/*
	 * We now have a TCP connection to the target
	 */

	/*
	 * The target should request no more than this number of
	 * blocks.
	 */
	max_seq = img_size / WAVR_BLOCK_SIZE;
	expected_seq = seq = 0;
	printf("Image size %lld bytes, About %d blocks.\n",
	    (uint64_t)st.st_size, (int)max_seq);

	/*
	 * Read and validate the sign-on.
	 */
	resp = get_response(fd);
	if (resp == NULL || strncmp(resp, "MAVERICK BOOTLOADER", 19) != 0) {
		local_errx(1, "Unexpected sign-on response '%s'", resp ? resp :
		    "[reply too long]");
	}

	/*
	 * Might as well let the target introduce itself to the user
	 */
	printf("%s\n", resp);

	/*
	 * Loop until the target tells us it's done
	 */
	for (retries = 0, seq = 0; retries < MAX_RETRIES; ) {
		char *p;

		/*
		 * Fetch the next reply
		 */
		resp = get_response(fd);
		if (resp == NULL) {
			local_errx(1, "\rTarget response exceeded length "
			    "limit.");
		}

		/*
		 * "COMPLETE" indicates the target is finished
		 */
		if (strcmp(resp, "COMPLETE\n") == 0) {
			resp = NULL;
			break;
		}

		/*
		 * We are now expecting the target to tell us what
		 * sequence/block number to send. Anything else is
		 * an error.
		 */
		if (strncmp(resp, "WANT ", 5) != 0)
			break;

		/*
		 * Ok so far. Decode the requested sequence number
		 */
		seq = strtoul(&resp[5], &p, 10);
		if (p == NULL || p == resp || (*p != '\n' && *p != '\0'))
			local_errx(1, "Invalid sequence number '%s'", resp);
		resp = NULL;

		/*
		 * Do not permit the target to request more blocks
		 * than we can provide. This generally indicates that
		 * the firmware image is corrupted.
		 */
		if (seq > max_seq) {
			local_errx(1, "Request block number (%d) is too big",
			    (int)seq);
		}

		/*
		 * The target is permitted to retry the previous block
		 * only. Anything earlier is an error.
		 */
		if (expected_seq && seq < (expected_seq - 1)) {
			local_errx(1, "Target is behind us! Expected %d, "
			    "got %d", (int)expected_seq, (int)seq);
		}

		/*
		 * The target is not permitted to get ahead of us
		 */
		if (seq > expected_seq) {
			local_errx(1, "Target is ahead of us. Expected %d, "
			    "got %d", (int)expected_seq, (int)seq);
		}

		/*
		 * Let the user know if the target requested a retry.
		 *
		 * Note: Our connection to the target is over TCP. In theory
		 * we should never need to retry blocks. However the WINC1500
		 * driver API, as provided by Atmel/Microchip, is flawed in a
		 * very specific way which can result data loss under certain
		 * circumstances. If anyone from Microchip is reading this I'm
		 * happy to discuss the problem!
		 */
		if (seq == (expected_seq - 1)) {
			printf("\nSequencing: re-sending block %d\n", (int)seq);
			retries += 1;
		} else {
			retries = 0;
		}

		/*
		 * Next block
		 */
		expected_seq = seq + 1;

		/*
		 * Display progress
		 */
		printf("\rProgress: %d%%", (int)((100.0 / (float)max_seq)*seq));
		fflush(stdout);

		/*
		 * Prepend the little-endian sequence number to the outgoing
		 * block.
		 */
		out_buff[0] = seq & 0xffu;
		out_buff[1] = (seq >> 8) & 0xffu;
		out_buff[2] = (seq >> 16) & 0xffu;
		out_buff[3] = (seq >> 24) & 0xffu;

		/*
		 * Copy the data to the out-going buffer
		 *
		 * Note: if you read the earlier comment regarding buggy
		 * WINC1500 driver/firmware, then you'll understand that
		 * sending the sequence number and image data in a single
		 * send(2) system call is appropriate. We could avoid this
		 * memcpy by using two send(2) syscalls but that increases
		 * the risk of packet loss.
		 */
		memcpy(&out_buff[4], &img[seq * WAVR_BLOCK_SIZE],
		    WAVR_BLOCK_SIZE);

		/*
		 * Send the block to the target
		 */
		if (send(fd, out_buff, sizeof(out_buff), 0) !=
		    sizeof(out_buff)) {
#ifdef _WIN32
			errstr = strerror(WSAGetLastError());
#endif
			local_err(1, "write block");
		}
	}

	closesocket(fd);
	free(img);

	if (resp) {
		printf("\rTarget response:\n%s\n", resp);
	} else
	if (retries == MAX_RETRIES) {
		printf("Giving up due to excessive retries for block %d\n",
		    (int)seq);
	} else {
		printf("\rProgress: 100%%\nSent %d blocks.\n", (int) seq + 1);
	}

#ifdef _WIN32
	WSACleanup();
#endif

	exit (0);
}
